function [classPrior,GMM,trainingClassificationRate,testClassificationRate,trainingCLL] = maxCllGMM(trainX,trainC,nIter,initGMM,initClassPrior,testX,testC,updatePrior,factor,useMex)
%[classPrior,GMM,trainingClassificationRate,testClassificationRate,trainingCLL] = maxCllGMM(trainX,trainC,nIter,initGMM,initClassPrior,testX,testC,factor,useMex)
%
%Train discriminative GMM using the conditional likelihood criterion.
%
%INPUT:
%trainX:                                    matrix: nTrain x nDim; Holds
%                                           all training samples.
%trainC:                                    vector: nTrain x 1; Holds all
%                                           training labels.
%nIter:                                     scalar; Number of iterations.
%initGMM:                                   cell array: nClass x 1; GMM
%                                           parameters for initialization.
%initClassPrior:                            vector: nClass x 1; Class prior
%                                           for initialization.
%testX:                                     matrix: nTest x nDim; Holds
%                                           all test samples for cross validation.
%testC:                                     vector: nTest x 1; Holds all
%                                           test labels.
%updatePrior:                               bool; If true, prior will be
%                                           updated too.
%factor:                                    scalar; Once all variances are positive, D1 will be multiplied by this factor.
%useMex:                                    bool; If true, the MATLAB MEX
%                                           implementation for faster
%                                           evaluation of the GMM will be used.
%
%OUTPUT:
%classPrior:                                vector: nClass x 1; The resulting
%                                           class prior.
%GMM:                                       cell array: nClass x 1; The
%                                           resulting GMM parameters.
%trainingClassificationRate:                vector: 1 x nIter; Percentage
%                                           of correctly classified samples on training set after each iteration.
%testClassificationRate:                    vector: 1 x nIter; Same for
%                                           test set.
%trainingCLL:                               vector: 1 x nIter; Conditional
%                                           log-likelihood on training set
%                                           after each iteration.



%author: Michael Wohlmayr, 19.02.2010
%v.0.1

%% PARAMETERS


%% Check input
[nTrain,nDim] = size(trainX);


%% 
%We assume that training class labels in trainC are within the range
%[1,max(trainC)].
%For some classes, it might happen that no training data is available. For
%this purpose, we create an internal class assignment which omits those
%classes without training data.
nExternalClass = max(trainC);
internalClass = zeros(nTrain,1);
newClassId = 1;
nInternalClass = length(unique(trainC));
externalClass = zeros(1,nInternalClass);
for kExternalClass = 1:nExternalClass
    classSampleIndex = (trainC==kExternalClass);
    if(any(classSampleIndex))
        internalClass(classSampleIndex) = newClassId; %assign every training sample an internal class ID
        externalClass(newClassId) = kExternalClass;
        newClassId = newClassId + 1;
    end
end

%%
%Discard samples from test data which belong to class that we omitted in
%the previous step. Assign internal class labels.
newC = zeros(1,size(testX,1));
for kClass = 1:nInternalClass
    newC(testC==externalClass(kClass)) = kClass;
end
testX = testX(newC~=0,:);
testC = newC(newC~=0);
nTest = size(testX,1);

%%
%Convert GMM parameters
mu = cell(1,nInternalClass); %mean
sigma = cell(1,nInternalClass); %diagonal covariance
w = cell(1,nInternalClass); %component priors
rho = zeros(1,nInternalClass); %class prior
nComp = zeros(1,nInternalClass); %number of components of each GMM

for kClass = 1:nInternalClass
    currentExternalClass = externalClass(kClass);
    nComp(kClass) = length(initGMM{currentExternalClass});
    mu{kClass}=zeros(nDim,nComp(kClass));
    sigma{kClass}=zeros(nDim,nComp(kClass));
    w{kClass}=zeros(1,nComp(kClass));
    rho(kClass) = initClassPrior(currentExternalClass);
    
    for kComp = 1:nComp(kClass)
        w{kClass}(1,kComp) = initGMM{currentExternalClass}{kComp}.prior;
        mu{kClass}(:,kComp) = initGMM{currentExternalClass}{kComp}.mean;
        sigma{kClass}(:,kComp) = diag(initGMM{currentExternalClass}{kComp}.cov);
    end
end

%% Pre-compute some quantities
u = zeros(nTrain,nInternalClass);
for kClass = 1:nInternalClass
    u(:,kClass) = internalClass==kClass;
end

%We need these two for partialRho
uNormalized = sum(u);
uNorm = sum(uNormalized);
uNormalized = uNormalized/uNorm;


trainingClassificationRate = zeros(1,nIter);
testClassificationRate = zeros(1,nIter);
trainingCLL = zeros(1,nIter);

for kIter = 1:nIter
    
    fprintf('Iteration %d\n',kIter);
    
    %% E-step
    
    %compute p(X|C) on training and test data, as well as p(k|X,C) (=gamma) on
    %training data.
    
    if(useMex)
        %use fast MEX-implementation
        [pX,gamma,pXTest] = evalGMMMEX(mu,sigma,w,trainX,testX);
        
        %check for infinite values in gamma
        for kClass = 1:nInternalClass
            actGamma = gamma{kClass};
            actGamma(~isfinite(actGamma)) = 0;
            gamma{kClass} = actGamma;
        end
        
    else
        %use MATLAB-implementation
        [pX,pXTest,gamma] = evalGMM(mu,sigma,w,trainX,testX,nComp,nInternalClass);
    end
    
    logRho = log(rho);
    
    %log p(X,C) on training set
    logPxc = pX + logRho(ones(nTrain,1),:);
    
    %log p(X,C) on test set
    logPxcTest = pXTest + logRho(ones(nTest,1),:);
    
    %r = p(C|X) (corresponds to w in pseudo code)
    rNorm = logsumexp(logPxc);
    r = logPxc - rNorm(:,ones(1,nInternalClass));
    r = exp(r);
    
    
    %We will need several times the term gamma*(u - r)
    s = cell(1,nInternalClass);
    for kClass = 1:nInternalClass
        part = u(:,kClass) - r(:,kClass);
        s{kClass} = gamma{kClass}.*part(:,ones(1,nComp(kClass)));
    end
    
    
    %% Classification rate
    [ignore,currentClassPrediction] = max(logPxc,[],2); %training set
    trainingClassificationRate(kIter) = sum(currentClassPrediction==internalClass)/nTrain*100;
    fprintf('Classification Rate on Training Set: %f\n',trainingClassificationRate(kIter));
    [ignore,currentClassPrediction] = max(logPxcTest,[],2); %test set
    testClassificationRate(kIter) = sum(currentClassPrediction==testC.')/nTest*100;
    fprintf('Classification Rate on Test Set: %f\n',testClassificationRate(kIter));
    clear currentClassPrediction;
    clear ignore;
    
    %% Conditional Log-Likelihood
    tmp = logPxc(logical(u)) - logsumexp(logPxc);
    trainingCLL(kIter) = sum(tmp);
    clear logPxc logPxcTest tmp;
    
    
    
    %% Constant D for EBW algorithm
    %D = inf;
    
    approxPartialAlpha = cell(1,nInternalClass);
    for kClass = 1:nInternalClass
        
        %approximate \frac{\partial \log D}{\partial \alpha_c^m} using two
        %fractions: Num1/Den1 - Num2/Den2
        part = gamma{kClass} .* u(:,kClass*ones(1,nComp(kClass)));
        Num1 = sum(part);
        Den1 = sum(Num1);
        
        
        part = gamma{kClass} .* r(:,kClass*ones(1,nComp(kClass)));
        Num2 = sum(part);
        Den2 = sum(Num2);
        
        approxPartialAlpha{kClass} = Num1/Den1-Num2/Den2;
        
        %D = min([D,approxPartialAlpha{kClass}.*w{kClass}]);
        
    end
    
    %approximate \frac{\partial \log D}{\partial \rho_c} using two
    %fractions: uNormalized - Num2/Den2
    
    Num1 = sum(r);
    Den1 = sum(Num1);
    
    approxPartialRho = uNormalized-Num1/Den1;
        
    clear Num1 Den1 Num2 Den2;
    
    %% M-Step
    D1 = 1;
    
    
    
    %mu and sigma
    %prepare all quantities for fast scanning of best D1
    partMuNum = cell(1,nInternalClass);
    partSigmaNum = cell(1,nInternalClass);
    Den = cell(1,nInternalClass);
    for kClass = 1:nInternalClass
        
        partMuNum{kClass} = zeros(nDim,nComp(kClass));
        partSigmaNum{kClass} = zeros(nDim,nComp(kClass));
        Den{kClass} = zeros(1,nComp(kClass));
        
        for kComp = 1:nComp(kClass)
            
            %mu
            partMuNum{kClass}(:,kComp) = sum(s{kClass}(:,kComp*ones(1,nDim)).*trainX).';
            
            %sigma
            partSigmaNum{kClass}(:,kComp) = sum(s{kClass}(:,kComp*ones(1,nDim)).*trainX.^2).';
            
        end
        %the denomiator is the same for both mu and sigma
        Den{kClass} = sum(s{kClass});
        
    end
    
    %Compute mu and sigma. As soon as a sigma gets negative, restart
    %computations with double D1.
    
    muNew = cell(1,nInternalClass);
    sigmaNew = cell(1,nInternalClass);
    negativeSigmaFound = 1;
    breakLoop = 0;
    while(negativeSigmaFound)
        
        negativeSigmaFound = 0;
        for kClass = 1:nInternalClass
            for kComp = 1:nComp(kClass)
                
                muOld = mu{kClass}(:,kComp);
                muOldSquare = muOld.^2;
                
                %mu
                muNew{kClass}(:,kComp) = (partMuNum{kClass}(:,kComp) + D1*muOld)/(Den{kClass}(kComp) + D1);
                
                %sigma
                sigmaNew{kClass}(:,kComp) = (partSigmaNum{kClass}(:,kComp) + D1*(sigma{kClass}(:,kComp)+muOldSquare))/(Den{kClass}(kComp) + D1) - muNew{kClass}(:,kComp).^2;
                
                %check if any variance is negative
                if(any(sigmaNew{kClass}(:,kComp)<0))
                    %restart the double loop with doubled D1.
                    D1 = D1 * 2;
                    negativeSigmaFound = 1;
                    breakLoop = 1;
                    break;
                end
                
            end
            
            if(breakLoop)
                breakLoop = 0;
                break;
            end
        end
        
    end
    
    %Again, multiply D1 with factor. If positivity check fails, fall back
    %to current D1.
    D1 = D1*factor;
  
    for kClass = 1:nInternalClass
        for kComp = 1:nComp(kClass)
            
            muOld = mu{kClass}(:,kComp);
            muOldSquare = muOld.^2;
            
            %mu
            muNew{kClass}(:,kComp) = (partMuNum{kClass}(:,kComp) + D1*muOld)/(Den{kClass}(kComp) + D1);
            
            %sigma
            sigmaNew{kClass}(:,kComp) = (partSigmaNum{kClass}(:,kComp) + D1*(sigma{kClass}(:,kComp)+muOldSquare))/(Den{kClass}(kComp) + D1) - muNew{kClass}(:,kComp).^2;
            
            %check if any variance is negative
            if(any(sigmaNew{kClass}(:,kComp)<0))
                %fall back to already confirmed D1
                D1 = D1/factor;
                breakLoop = 1;
                break;
            end
            if(breakLoop)
                breakLoop = 0;
                break;
            end
            
        end
    end
    
    
    for kClass = 1:nInternalClass
        for kComp = 1:nComp(kClass)
            
            muOld = mu{kClass}(:,kComp);
            muOldSquare = muOld.^2;
            
            %mu
            muNew{kClass}(:,kComp) = (partMuNum{kClass}(:,kComp) + D1*muOld)/(Den{kClass}(kComp) + D1);
            
            %sigma
            sigmaNew{kClass}(:,kComp) = (partSigmaNum{kClass}(:,kComp) + D1*(sigma{kClass}(:,kComp)+muOldSquare))/(Den{kClass}(kComp) + D1) - muNew{kClass}(:,kComp).^2;
            
        end
    end
    
    
    clear partMuNum partSigmaNum Den;
    
    %rho
    if(updatePrior)
        rho = rho.*(approxPartialRho + D1);
        rho = rho./sum(rho);
    end
    
    % w (corresponds to alpha in pseudo-code)
    for kClass = 1:nInternalClass
        if(nComp(kClass)==1)
            w{kClass} = 1;
        else
            newW = w{kClass}.*(approxPartialAlpha{kClass} + D1);
            w{kClass} = newW/sum(newW);
        end
    end
    clear approxPartialAlpha;
    
    mu = muNew;
    sigma = sigmaNew;
    
    
end

classPrior = rho;
GMM = initGMM;
%write modified parameters to GMM
for kClass = 1:nInternalClass
    for kComp = 1:nComp(kClass)
        GMM{externalClass(kClass)}{kComp}.prior = w{kClass}(kComp);
        GMM{externalClass(kClass)}{kComp}.mean = mu{kClass}(:,kComp);
        GMM{externalClass(kClass)}{kComp}.cov = diag(sigma{kClass}(:,kComp));
    end
end


function [pX,pXTest,gamma] = evalGMM(mu,sigma,w,trainX,testX,nComp,nInternalClass)
%Given a set of |C| GMMs, compute log p(X|C) on training and test data, as well
%as p(k|X,C) for the training data (where k is the component index of the
%C-th GMM).
%
%INPUT:
%mu:                cell array; mu{c}(:,k) has -kth mean of c-th GMM.
%sigma:             cell array; sigma{c}(:,k) has -kth diagonal covariance of c-th GMM.
%w:                 cell array; w{c}(k) has k-th weight of c-th GMM.
%trainX:            matrix; trainX(n,:) has n-th training sample.
%testX:             matrix; trainX(n,:) has n-th test sample.
%nComp:             vector; nComp(c) has number of components of c-th GMM.
%nInternalClass     scalar; number of GMMs
%
%OUTPUT:
%pX:                matrix; pX(n,c) gives log p(x^n|C==c) for training data.
%pXTest:            matrix; pX(n,c) gives log p(x^n|C==c) for test data.
%gamma:             cell array; gamma{c}(n,k) gives component posterior
%                   p(k|x^n,C==c).
%
%evalDataMEX is the equivalent implementation of this function using MEX.

%author: Michael Wohlmayr, 19.02.2010

[nTrain,nDim] = size(trainX);
nTest = size(testX,1);
const1 = nDim*0.5*log(2*pi); %for computation of normal distribution
gamma = cell(1,nInternalClass);
pX = zeros(nTrain,nInternalClass); %p(x^n|\theta_c) on training set
pXTest = zeros(nTest,nInternalClass); %p(x^n|\theta_c) on test set
for kClass = 1:nInternalClass
    
    gamma{kClass} = zeros(nTrain,nComp(kClass));
    logW = log(w{kClass});
    
    actMuMatrix = mu{kClass}.';
    actStdMatrix = sqrt(sigma{kClass}).';
    logPXallComp = zeros(nTrain,nComp(kClass));
    logPXTestAllComp = zeros(nTest,nComp(kClass));
    for kComp = 1:nComp(kClass)
        
        actMu = actMuMatrix(kComp,:);
        actStd = actStdMatrix(kComp,:);
        logActStd = log(actStd);
        
        %probability of trainX for all GMM components
        Z = (trainX-actMu(ones(1,nTrain),:))./actStd(ones(1,nTrain),:); %standardize: Z = (X-mu)/sigma
        Z = Z.^2;
        logPXallComp(:,kComp) = -const1-sum(0.5*Z+logActStd(ones(1,nTrain),:),2);
        
        
        %probability of testX for all GMM components
        Z = (testX-actMu(ones(1,nTest),:))./actStd(ones(1,nTest),:); %standardize: Z = (X-mu)/sigma
        Z = Z.^2;
        logPXTestAllComp(:,kComp) = -const1-sum(0.5*Z+logActStd(ones(1,nTest),:),2);
        
    end
    
    %gamma and p(x^n|\theta_c) on training set
    if(nComp(kClass)==1)
        gamma{kClass} = ones(nTrain,1);
        pX(:,kClass) = logPXallComp + logW(ones(nTrain,1));
    else
        actGamma = logPXallComp + logW(ones(nTrain,1),:);
        normGamma = logsumexp(actGamma);
        actGamma = actGamma - normGamma(:,ones(1,nComp(kClass)));
        gamma{kClass} = exp(actGamma);
        
        %p(x^n|\theta_c) for training set
        pX(:,kClass) = normGamma;
    end
    
    
    
    %p(x^n|\theta_c) for test set
    pXTest(:,kClass) = logsumexp(logPXTestAllComp + logW(ones(nTest,1),:));
    
end

function y = logsumexp(x)
%logsumexp over rows of x in numerically stable way.

nCol = size(x,2);
m = max(x,[],2);
mr = m(:,ones(1,nCol));
y = m + log(sum(exp(x - mr),2));